"""Session interfaces for interacting with Aquarium.

Session interfaces are created by an AqSession instance and use an AqHTTP
instance to make http requests to Aquarium.

Generally, Trident users should be unable to directly access AqHTTP method to
prevent potentially damaging requests or revealing sensitive information.
Session interfaces define a particular way an AqHTTP is used.

Interfaces are roughly grouped into:

(1) ModelInterface - for finding and retrieving Aquarium models and information
(2) CreateInterface - for creating new objects (e.g. new Samples) in Aquarium
(3) UpdateInterface - for updating objects in Aquarium

Example:
    session1 = AqSession(login, password, aquarium_url)
    session1.User
    # => a ModelInterface instance

    session1.User.find(1)
    # use model interface with User to find User(id=1)

    session1.aqhttp.post(...)
    # not allowed

    session1.create
    # => a CreateInterface

    session1.create.samples(list_of_samples)
    # creates samples from a list by calling method CreateInterface.samples
"""
import json
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Union

from inflection import pluralize
from inflection import underscore

from .exceptions import TridentRequestError
from .utils import url_build
from pydent.marshaller.registry import ModelRegistry


class SessionInterface:
    """Generic session interface.

    Trident users should only be able to make requests through a
    SessionInterface to avoid making arbitrary and potentially damaging
    http requests.
    """

    __slots__ = ["aqhttp", "session"]

    def __init__(self, aqhttp, session):
        """Initializer for SessionInterface.

        :param aqhttp: aqhttp instance for this interface
        :type aqhttp: AqHTTP
        :param session: session instance for this interface
        :type session: AqSession
        """
        self.aqhttp = aqhttp
        self.session = session


class CRUDInterface(SessionInterface):
    """Create, Update, Delete interface.

    Methods for communicating with Aquarium's JSON controller.
    """

    __slots__ = ["aqhttp", "session"]

    # def _validate_model_id(self, model):
    #     if not hasattr(model, 'id') or model.id is None:
    #         raise TridentRequestError("Model {} has no id.".format(model))

    def _model_controller(
        self,
        method: str,
        table: str,
        model_id: Union[str, int, None],
        data: Union[None, dict],
        params=None,
    ) -> dict:
        """Method for create, updating, and deleting models.

        :param method: Request method name (one of 'put', 'post', 'delete'
        :param table: Table name of model (e.g. 'samples' or 'data_associations')
        :param model_id: Optional model_id (not required for 'post')
        :param data: data
        :param params
        :return: json formatted server response
        :rtype: dict
        """
        if model_id:
            url = "{}/{}.json".format(table, model_id)
        else:
            url = "{}.json".format(table)

        result = self.aqhttp.request(method, url, json=data, params=params)
        return result

    def model_create(self, table, data, params=None):
        return self._model_controller("post", table, None, data, params)

    def model_update(self, table, model_id, data, params=None):
        return self._model_controller("put", table, model_id, data, params)

    def model_delete(self, table, model_id, params=None):
        if model_id is None:
            return
        return self._model_controller("delete", table, model_id, None, params)

    def _json_controller(
        self,
        method,
        model_name,
        model_data,
        record_methods: List[str] = None,
        record_getters: List[str] = None,
    ):
        """Method for create, updating, and deleting models using Aquariums
        JSON controller.

        :param method: Method name (e.g. "save", "delete")
        :type method: basestring
        :param model_name: Model name
        :type model_name: basestring
        :param model_data: Additional model and method data
        :type model_data: dict
        :param record_methods: Optional 'record_methods' key
        :type record_methods: dict
        :param record_getters: Optional 'record_getters' key
        :type record_getters: dict
        :return: json formatter response
        :rtype: basestring
        """

        if record_methods is None:
            record_methods = {}
        if record_getters is None:
            record_getters = {}

        data = {
            "model": {
                "model": model_name,
                "record_methods": record_methods,
                "record_getters": record_getters,
            }
        }

        if model_data:
            data.update(model_data)

        if method:
            url = "json/" + method
        else:
            url = "json"

        try:
            post_response = self.aqhttp.post(url, json_data=data)
            return post_response
        except TridentRequestError as err:
            raise err

        # TODO: is this code necessary?
        # result = self.aqhttp.post("json" + method, json_data=data)
        # return result

    def json_delete(
        self,
        model_name,
        model_data,
        record_methods: List[str] = None,
        record_getters: List[str] = None,
    ):
        return self._json_controller(
            "delete", model_name, model_data, record_methods, record_getters
        )

    def json_save(
        self,
        model_name,
        model_data,
        record_methods: List[str] = None,
        record_getters: List[str] = None,
    ) -> dict:
        return self._json_controller(
            "save", model_name, model_data, record_methods, record_getters
        )

    def json_post(
        self,
        model_name,
        model_data,
        record_methods: List[str] = None,
        record_getters: List[str] = None,
    ):
        return self._json_controller(
            None, model_name, model_data, record_methods, record_getters
        )


class UtilityInterface(CRUDInterface):
    """Miscellaneous and specialized requests for creating, updating, etc."""

    __slots__ = ["aqhttp", "session"]

    ##############################
    # Create
    ##############################

    # TODO: have ability to save new properties
    def create_samples(self, samples):
        json_data = [s.dump(include=("field_values",)) for s in samples]

        updated_sample_data = self.aqhttp.post(
            "browser/create_samples", {"samples": json_data}
        )
        updated_samples = updated_sample_data["samples"]
        assert len(samples) == len(updated_samples)
        for s, data in zip(samples, updated_samples):
            s.reload(data)
        return samples

    def create_items(self, items):
        def sid(i):
            if i.sample_id:
                return i.sample_id
            return i.sample.id

        def otid(i):
            if i.object_type_id:
                return i.object_type_id
            return i.object_type.id

        return [
            self.aqhttp.get("items/make/{}/{}".format(sid(i), otid(i))) for i in items
        ]

    def create_upload(self, upload):
        files = {"file": upload.file}
        result = self.aqhttp.post(
            "krill/upload?job={}".format(upload.job_id), files=files
        )
        upload.reload(result)
        return upload

    def create_data_association(self, model_inst, key, value, upload=None):
        upload_id = None
        if not model_inst.id:
            raise ValueError(
                "Cannot create DataAssociation because model has no id ("
                "it probably is not saved on the server)"
            )
        if upload is not None:
            upload_id = upload.id
        data = {
            "parent_id": model_inst.id,
            "key": str(key),
            "object": json.dumps({str(key): value}),
            "parent_class": model_inst.__class__.__name__,
            "upload_id": upload_id,
        }
        result = self.json_save("DataAssociation", data)

        data_association = model_inst.session.DataAssociation.find(result["id"])
        if data_association.id not in [da.id for da in model_inst.data_associations]:
            model_inst.data_associations.append(data_association)
            return data_association
        else:
            for da in model_inst.data_associations:
                if da.id == data_association.id:
                    return da

    ##############################
    # Save/Update
    ##############################

    def update_code(self, code):
        """Updates code for a operation_type."""
        controller = underscore(pluralize(code.parent_class))

        code_data = {"id": code.parent_id, "name": code.name, "content": code.content}
        result = self.aqhttp.post(url_build(controller, "code"), code_data)
        if "id" in result:
            code.id = result["id"]
            code.parent_id = result["parent_id"]
            code.updated_at = result["updated_at"]
        else:
            raise TridentRequestError(
                "Unable to update code object {}".format(code_data), result
            )

    ##############################
    # Misc
    ##############################

    def move_item(self, item, location):
        assert item.id
        url = "items/move/{item_id}?location={location}".format(
            item_id=item.id, location=location
        )
        result = self.aqhttp.get(url)
        if "error" in result or "errors" in result:
            raise TridentRequestError(str(result))
        item.location = location
        return item

    def store_item(self, item):
        assert item.id
        self.aqhttp.get("items/store/{item_id}")
        item.refresh()

    def estimate_plan_cost(self, plan):
        """Estimates the plan cost."""
        result = self.aqhttp.post("launcher/estimate", {"id": plan.id})
        total = 0
        for operation in plan.operations:
            for cost in result["costs"]:
                if cost["id"] == operation.id:
                    operation.cost = cost
                    total += (
                        cost["labor"] * cost["labor_rate"] + cost["materials"]
                    ) * cost["markup_rate"]
        return total

    def step_plan(self, plan_id):
        self.aqhttp.get("operations/step?plan_id={}".format(plan_id))
        return None

    def replan(self, plan_id):
        """Copies a plan."""
        result = self.aqhttp.get("plans/replan/{plan_id}".format(plan_id=plan_id))
        plan_copy = self.session.Plan.load(result)
        return plan_copy

    def start_job(self, job_id):
        result = self.aqhttp.get("krill/start?job={job_id}".format(job_id=job_id))
        return result

    def batch_operations(self, operation_ids):
        """Batches operations from a list of operation_ids."""
        self.aqhttp.post("operations/batch", json_data=operation_ids)

    def unbatch_operations(self, operation_ids):
        """Unbatches operations from a list of operation_ids."""
        self.aqhttp.post("operations/batch", json_data=operation_ids)

    def set_operation_status(self, operation_id, status):
        """Sets an operation's status."""
        msg = "operations/{oid}/status/{status}".format(oid=operation_id, status=status)
        self.aqhttp.get(msg)

    def job_debug(self, job_id):
        """Runs debug on a job with id=job_id."""
        self.aqhttp.get("krill/debug/{jid}".format(jid=job_id))

    def submit_plan(self, plan, user, budget):
        """Submits a plan."""
        user_query = "&user_id=" + str(user.id)
        budget_query = "?budget_id=" + str(budget.id)
        result = self.aqhttp.get(
            "plans/start/" + str(plan.id) + budget_query + user_query
        )
        return result

    def compatible_items(self, sample_id, object_type_id):
        """Find items compatible with the field value."""
        result = self.aqhttp.post(
            "json/items", {"sid": sample_id, "oid": object_type_id}
        )
        items = []
        for element in result:
            print(element)
        return items


class QueryInterfaceABC(ABC):
    """Interface that is used by models to find other models."""

    @abstractmethod
    def find(self, mid):
        pass

    @abstractmethod
    def find_by_name(self, name):
        pass

    @abstractmethod
    def all(self):
        pass

    @abstractmethod
    def where(self, query, **kwargs):
        pass

    @abstractmethod
    def first(self):
        pass

    @abstractmethod
    def last(self):
        pass

    @abstractmethod
    def one(self):
        pass

    @abstractmethod
    def model_name(self):
        """Alias for self.model.__name__"""
        pass


class QueryInterface(SessionInterface, QueryInterfaceABC):
    """Makes requests using AqHTTP that are model specific.

    Establishes a connection between a session object and an Aquarium
    model.
    """

    __slots__ = ["aqhttp", "session", "model", "__dict__"]
    MERGE = "query_hook"
    DEFAULT_OFFSET = -1
    DEFAULT_REVERSE = False
    DEFAULT_LIMIT = -1

    def __init__(self, model_name, aqhttp, session):
        """Instantiates a new model interface. Uses aqhttp to make requests,
        and deserializes response to models.

        :param model_name: Model name (e.g. 'Sample' or 'FieldValue')
        :type model_name: basestring
        :param aqhttp: the AqHTTP instance
        :type aqhttp: AqHTTP
        :param session:
        :type session:
        """
        super().__init__(aqhttp, session)
        self.crud = CRUDInterface(aqhttp, session)
        self.model = ModelRegistry.get_model(model_name)
        self._do_load = True

    @property
    def model_name(self):
        return self.model.__name__

    def _prepost_query_hook(self, query):
        """Method for modifying the query before posting."""

        additional_query = {}
        if hasattr(self.model, self.MERGE):
            additional_query = getattr(self.model, self.MERGE)

        # update and merge
        for k, v in additional_query.items():
            if k not in query:
                query[k] = v
            elif isinstance(query[k], list):
                if isinstance(v, list):
                    query[k] += v
                else:
                    query[k].append(v)
            elif isinstance(query[k], dict):
                if isinstance(v, dict):
                    query[k].update(v)
                else:
                    raise ValueError(
                        "Cannot update query {} with {}:{}".format(query, k, v)
                    )
        return query

    def _post_json(self, data):
        """Posts a json request to session for this interface.

        Attaches raw json and this session instance to the models it
        retrieves.
        """
        data_dict = {"model": self.model_name}
        data_dict = self._prepost_query_hook(data_dict)
        data_dict.update({k: v for k, v in data.items() if v})

        try:
            post_response = self.crud.json_post(self.model_name, data_dict)
        except TridentRequestError as err:
            if err.response.status_code == 422:
                return None
            else:
                raise err

        if post_response is not None and self._do_load:
            return self.load(post_response)
        return post_response

    def load(self, post_response):
        """Loads model instance(s) from data.

        Model instances will be of class defined by self.model. If data
        is a list, will return a list of model instances.
        """
        models = self.model.load_from(post_response, self.session)
        return models

    def get(self, path):
        """Makes a generic get request."""
        try:
            response = self.aqhttp.get(path)
        except TridentRequestError as err:
            if err.response.status_code == 404:
                return None
            raise err
        return self.load(response)

    def find(self, model_id, include=None, opts: dict = None):
        """Finds model by id."""
        if model_id is None:
            raise ValueError("model_id in 'find' cannot be None")
        if model_id == 0:
            return None
        return self._post_json({"id": model_id, "include": include, "options": opts})

    def find_by_name(self, name, include=None, opts: dict = None):
        """Finds model by name."""
        if name is None:
            raise ValueError("name in 'find_by_name' cannot be None")
        if name.strip() == "":
            return None
        return self._post_json(
            {
                "method": "find_by_name",
                "arguments": [name],
                "include": include,
                "options": opts,
            }
        )

    def array_query(self, method, args, rest=None, include=None, opts: dict = None):
        """Finds models based on a query."""
        if opts is None:
            opts = {}
        options = {
            "offset": self.DEFAULT_OFFSET,
            "limit": self.DEFAULT_LIMIT,
            "reverse": self.DEFAULT_REVERSE,
        }
        options.update(opts)
        if options.get("limit", None) == 0:
            return []
        if args is None:
            args = []
        query = {
            "model": self.model.__name__,
            "method": method,
            "arguments": args,
            "include": include,
            "options": options,
        }
        if rest:
            query.update(rest)
        res = self._post_json(query)
        if res is None:
            return []
        return res

    def all(self, methods: List[str] = None, include=None, opts: dict = None):
        """Finds all models.

        :param methods:
        :param include:
        :type include:
        :param opts: additional options ("offset", "limit", "reverse", etc.)
        :type opts: dict
        :return:
        :rtype:
        """

        if opts is None:
            opts = {}
        addopts = opts.pop("opts", dict())
        opts.update(addopts)
        options = {"offset": self.DEFAULT_OFFSET, "reverse": self.DEFAULT_REVERSE}
        options.update(opts)
        return self.array_query(
            method="all", args=None, rest=None, include=include, opts=options
        )

    def where(
        self,
        criteria: dict,
        methods: List[str] = None,
        include: List[str] = None,
        opts: dict = None,
    ):
        """Performs a query for models.

        :param criteria: query to find models
        :type criteria: dict
        :param methods: server side methods to implement
        :type methods: list
        :param opts: additional options ("offset", "limit", "reverse", etc.)
        :type opts: dict
        :param include:
        :return: list of models
        :rtype: list
        """
        if opts is None:
            opts = dict()
        rest = {}
        if methods is not None:
            rest = {"methods": methods}
        return self.array_query(
            method="where", args=criteria, rest=rest, include=include, opts=opts
        )

    # TODO: Refactor 'last' so query is an argument, not part of kwargs
    def last(
        self, num: int = None, query: dict = None, include=None, opts: dict = None
    ):
        """Find the last added models.

        :param num: number of models to return. If not provided, assumes 1
        :type num: int
        :param query: additional query to find models
        :type query: dict
        :param opts: additional options ("offset", "limit", "reverse", etc.)
        :type opts: dict
        :return: list of models
        :rtype: list
        """
        if query is None:
            query = dict()
        if num is None:
            num = 1
        if opts is None:
            opts = dict()
        opts.update(dict(limit=num, reverse=True))
        return self.where(query, include=include, opts=opts)

    # TODO: Refactor 'first' so query is an argument, not part of kwargs
    def first(
        self, num: int = None, query: dict = None, include=None, opts: dict = None
    ):
        """Find the first added models.

        :param num: number of models to return. If not provided, assumes 1
        :type num: int
        :param query: additional query to find models
        :type query: dict
        :param include:
        :param opts: additional options ("offset", "limit", "reverse", etc.)
        :type opts: dict
        :return: list of models
        :rtype: list
        """
        if query is None:
            query = dict()
        if num is None:
            num = 1
        if opts is None:
            opts = dict()
        opts.update(dict(limit=num, reverse=False))
        return self.where(query, include=include, opts=opts)

    # TODO: Refactor 'one' so query is an argument, not part of kwargs
    def one(
        self, query: dict = None, first: bool = False, include=None, opts: dict = None
    ):
        """Return one model. Returns the last model by default. Returns None if
        no model is found.

        :param first: whether to return the first model (default: False
        :type first: bool
        :return: model
        :rtype: ModelBase
        """
        if not first:
            res = self.last(1, query=query, include=include, opts=opts)
        else:
            res = self.first(1, query=query, include=include, opts=opts)
        if not res:
            return None
        else:
            return res[0]

    # TODO: implement 'patch' or 'update'? Would this be too dangerous?
    # def patch(self, model_id, json_data):
    #     """
    #     Makes a patch or update from json_data
    #     """
    #     models_name = pluralize(underscore(self.model_name))
    #     result = self.aqhttp.put('{models}/{model_id}'.format(
    #         model_id=model_id, models=models_name), json_data=json_data)
    #     return result

    def new(self, *args, **kwargs):
        """Creates a new model instance.

        Attach a session by calling __new__ with session kwargs.
        """
        instance = self.model.__new__(self.model, *args, session=self.session, **kwargs)
        self.model.__init__(instance, *args, **kwargs)
        return instance

    def __call__(self, *args, **kwargs):
        return self.new(*args, **kwargs)


class BrowserInterface(SessionInterface, QueryInterfaceABC):
    __slots__ = ["aqhttp", "session", "model", "__dict__"]
    MERGE = ["methods"]
    DEFAULT_OFFSET = -1
    DEFAULT_REVERSE = False
    DEFAULT_LIMIT = -1

    # TODO: browser should use standard session
    def __init__(self, model_name, aqhttp, session):
        super().__init__(aqhttp, session)
        self.crud = CRUDInterface(aqhttp, session)
        self.model = ModelRegistry.get_model(model_name)
        self._do_load = True
        self._preload_field_values = None

    @property
    def model_name(self):
        return self.model.__name__

    @property
    def browser(self):
        return self.session.browser

    def find(self, model_id):
        return self.browser.find(model_id, model_class=self.model_name)

    def find_by_name(self, name):
        return self.browser.find_by_name(name, model_class=self.model_name)

    def where(self, criteria, methods: List[str] = None, opts: bool = None):
        return self.browser.where(
            criteria, model_class=self.model_name, methods=methods, opts=opts
        )

    def one(self, query: dict = None, first: bool = False, opts: bool = None):
        return self.browser.one(model_class=self.model_name, query=query, opts=opts)

    def first(self, num: int = 1, query: dict = None, opts: bool = None):
        return self.browser.first(num, model_class=self.model_name, query=query)

    def last(self, num: int = 1, query: dict = None, opts: bool = None):
        return self.browser.last(num, model_class=self.model_name, query=query)

    def new(self, *args, **kwargs):
        instance = self.model.__new__(self.model, *args, session=self.session, **kwargs)
        self.model.__init__(instance, *args, **kwargs)
        self.browser.update_cache([instance])
        return instance

    def all(self, opts: bool = None):
        return self.browser.all(model_class=self.model_name, opts=opts)

    # TODO: load_from using new session
    def load(self, post_response: dict) -> List["ModelBase"]:
        """Loads model instance(s) from data.

        Model instances will be of class defined by self.model. If data
        is a list, will return a list of model instances.
        """
        models = self.model.load_from(post_response, self.session)
        return models



"""
Browser (:mod:`pydent.browser`)
=================================

.. versionadded:: 0.1
    Browser class created

.. currentmodule:: pydent.browser

Browser class for searching and cacheing results.
"""
import re
from collections import OrderedDict
from difflib import get_close_matches
from pprint import pformat
from typing import Callable
from typing import Dict
from typing import List
from typing import Union

import networkx as nx

from pydent import models as pydent_models
from pydent.base import ModelBase
from pydent.exceptions import ForbiddenRequestError
from pydent.exceptions import TridentBaseException
from pydent.interfaces import QueryInterface
from pydent.interfaces import QueryInterfaceABC
from pydent.marshaller import ModelRegistry
from pydent.models import Sample
from pydent.relationships import BaseRelationship
from pydent.sessionabc import SessionABC
from pydent.utils import logger
from pydent.utils.logging_helpers import did_you_mean

# TODO: browser documentation
# TODO: examples in sphinx
# TODO: methods to help pull relevant data from plans (user specifies types of data to
#       pull, and trident should pull and cache in the most efficient way possible)


class BrowserException(TridentBaseException):
    """Generic browser exception."""


class Browser(QueryInterfaceABC):
    """A class for browsing models and Aquarium inventory."""

    # TODO: ability to block model callbacks to enforce cache

    INTERFACE_CLASS = QueryInterface
    ACCEPTED_GET_RELATION_TYPES = [
        "HasOne",
        "HasMany",
        "HasManyThrough",
        "HasManyGeneric",
    ]

    def __init__(self, session: SessionABC, inherit_models: bool = False):
        """Instantiates a new browser from a AqSession instance.

        .. versionchanged:: 0.1.5a7
            'inherit_models' argument will inherit the sessions model_cache
            (default: False)

        :param session: a session instance
        :param inherit_models: if True, the browser will inherit the cache in the
            provided session's browser model_cache
        :type session: SessionABC
        """
        self.session = session
        self._list_models_fxn = self.sample_list
        self.use_cache = True
        self.model = Sample
        self.model_list_cache = {}
        self.model_cache = {}
        self.log = logger(name="Browser@{}".format(session.url))
        if session.browser and inherit_models:
            self.update_cache(session.browser.models)

    @property
    def model_name(self):
        return self.model.__name__

    # TODO: change session interface (find, where, etc.) to use cache IF use_
    #       cache = True
    # TODO: where and find queries can sort through models much more quickly than
    #       Aquarium, but can fallback to Aq

    @property
    def models(self):
        models = []
        for v in self.model_cache.values():
            models += list(v.values())
        return models

    def set_model(self, model_name):
        """Sets the default model of this browser."""
        self.model = ModelRegistry.get_model(model_name)
        if model_name == "Sample":
            self._list_models_fxn = self.sample_list
        else:
            self._list_models_fxn = self._generic_list_models

    def interface(self, model_class=None):
        """Returns a new model query interface.

        :param model_class:
        :type model_class: basestring
        :return: Interface
        :rtype: QueryInterface
        """
        if model_class is None:
            model_class = self.model_name
        return self.session.model_interface(
            model_class, interface_class=self.INTERFACE_CLASS
        )

    def _generic_list_models(self, **query):
        models = self.where(query, self.model_name)
        return ["{}: {}".format(m.id, m.name) for m in models]

    def sample_list(self, sample_type_id=None):
        """Returns a sample list."""
        path = "sample_list"
        if sample_type_id is not None:
            path += "/" + str(sample_type_id)
        return self.session.utils.aqhttp.get(path)

    def clear(self):
        """Clears the model cache."""
        self.model_list_cache = {}
        self.model_cache = {}

    def list_models(self, *args, **kwargs):
        def get_models():
            return self._list_models_fxn(*args, **kwargs)

        if self.use_cache:
            models_cache = self.model_list_cache.get("models", {})
            if self.model_name in models_cache:
                model_list = models_cache[self.model_name]
            else:
                model_list = get_models()
        else:
            model_list = get_models()
        self.model_list_cache.setdefault("models", {})
        self.model_list_cache["models"][self.model_name] = model_list[:]
        return model_list

    def where(
        self, query, model_class=None, primary_key="id", sample_type=None, **kwargs
    ):
        """Perform a 'where' query. If models are found in the browser cache,
        those are returned, else new http queries are made to find the models.

        :param query: query as a dictionary
        :param model_class: model class to use (str)
        :param primary_key: which primary key to use (default: 'id')
        :param sample_type: optional sample_type short cut for finding samples
        :param kwargs: other kwargs
        :return: returned model list
        """
        if model_class is None:
            model_class = self.model_name
        if sample_type is not None:
            sample_type_id = self.find_by_name(sample_type, "SampleType").id
            query.update({"sample_type_id": sample_type_id})
        if self.use_cache:
            return self.cached_where(query, model_class, primary_key=primary_key)
        return self.interface(model_class).where(query, **kwargs)

    def __query_helper(
        self,
        fname,
        query,
        model_class,
        sample_type=None,
        opts=None,
        params=None,
        as_single=False,
    ):
        """Builds a custom query for the browser.

        :param fname: the function name
        :type fname: basestring
        :param query: the query
        :type query: dict
        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param sample_type: optional sample_type name
        :type sample_type: basestring
        :param opts: options to send to the function
        :type opts: dict
        :param params: additionaly keyword arguments to send to the function
        :type params: dict
        :param as_single: if True, will return the first model of the array or None if
            array is empty
        :type as_single: bool
        :return: Aquarium model or list of Aquarium models
        :rtype: ModelBase | list
        """
        if model_class is None:
            model_class = self.model_name
        if query is None:
            query = dict()
        if params is None:
            params = dict()
        if sample_type is not None:
            query.update(
                {"sample_type_id": self.find_by_name(sample_type, "SampleType").id}
            )
        interface = self.interface(model_class)
        fxn = getattr(interface, fname)
        if fname == "all":
            models = fxn(opts=opts, **params)
        else:
            models = fxn(query=query, opts=opts, **params)
        if as_single:
            models = [models]
        return self.update_cache(models).get(model_class, [])

    def one(self, model_class=None, sample_type=None, query=None, opts=None):
        """Finds one instance of a model (or returns None)

        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param sample_type: optional sample_type name
        :type sample_type: basestring
        :param query: additional query to filter models
        :type query: dict
        :param opts: additional options
        :type opts: dict
        :return:
        :rtype:
        """
        models = self.__query_helper(
            "one", query, model_class, sample_type, opts=opts, as_single=True
        )
        if not models:
            return None
        return models[0]

    def last(self, num=1, model_class=None, sample_type=None, query=None):
        """Finds last models. Will NOT return cached models.

        :param num: number of models to return
        :type num: int
        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param sample_type: optional sample_type name
        :type sample_type: basestring
        :param query: additional query to filter models
        :type query: dict
        :return:
        :rtype:
        """
        return self.__query_helper(
            "last", query, model_class, sample_type, params=dict(num=num)
        )

    def first(self, num=1, model_class=None, sample_type=None, query=None):
        """Finds first models. Will NOT return cached models.

        :param num: number of models to return
        :type num: int
        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param sample_type: optional sample_type name
        :type sample_type: basestring
        :param query: additional query to filter models
        :type query: dict
        :return:
        :rtype:
        """
        return self.__query_helper(
            "first", query, model_class, sample_type, params=dict(num=num)
        )

    def find(self, model_id, model_class=None):
        """Finds a model by id. Will returned cached model if possible.

        :param model_id: model_id
        :type model_id: int
        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :return:
        :rtype:
        """
        if model_class is None:
            model_class = self.model_name
        if self.use_cache:
            return self.cached_find(model_class, model_id)
        return self.interface(model_class).find(model_id)

    def find_by_name(self, name, model_class=None, primary_key="id"):
        """Find model by name. Will return cached model if possible.

        :param name: name of the model
        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param primary_key:
        :return:
        """
        models = self.where({"name": name}, model_class, primary_key=primary_key)
        if not models:
            return None
        return models[0]

    def all(self, model_class=None, opts=None):
        """Return all models of a model_class.

        :param model_class: the name of the model class (e.g. "Sample")
        :type model_class: basestring
        :param opts:
        :type opts:
        :return:
        :rtype:
        """
        return self.__query_helper("all", query={}, model_class=model_class, opts=opts)

    @staticmethod
    def _match_query(query, model_dict):
        """Matches a query against a model dictionary.

        :param query: query dictionary
        :type query: dict
        :param model_dict: model dictionary (e.g. model.__dict__)
        :type model_dict: dict
        :return: whether the model matches the query
        :rtype: bool
        """
        for query_key in query:
            if query_key not in model_dict:
                return False
            query_val = query[query_key]
            model_val = model_dict[query_key]
            if isinstance(query_val, list):
                if model_val not in query_val:
                    return False
            elif model_val != query_val:
                return False
        return {k: model_dict[k] for k in query}

    @classmethod
    def _find_matches(cls, query, models):
        found = []
        found_queries = []
        for m in models:
            match = cls._match_query(query, m._get_data())
            if match:
                found.append(m)
                found_queries.append(match)
        return found, found_queries

    def update_cache(
        self,
        models: List[ModelBase],
        recursive: bool = True,
        update_session: bool = False,
    ) -> Dict[str, List[ModelBase]]:
        """Update the browser's model cache from a list of models.

        .. versionchanged:: 0.1.5a8
            Added 'update_session' kwarg which will change the model's
            session to the browser's session.

        .. warning::
            If `update_session=True`, the model's session
            will change and this may have unintended consequences.

        :param models: list of models
        :param recursive: if True, recursively collect all models contained in
            the relationships and use those to update the cache as well.
        :param update_session: if True, attach the browser's session to the models
        :return: the updated dictionary broken down by model class name
        """
        assert isinstance(models, list)
        if recursive:
            memo = {}
            ModelBase._flatten_deserialized_data(models, memo)
            models = list(memo.values())
        if update_session:
            for m in models:
                m._session = self.session
        return self._group_models_and_update_cache(models)

    # TODO: do we really want to simply overwrite the dictionary or update the models?
    def _update_model_cache_helper(self, modelname, modeldict):
        """Updates the browser's model cache with models from the provided
        model dict."""
        self.log.info(
            "CACHE updated cached with {} {} models".format(len(modeldict), modelname)
        )
        self.model_cache.setdefault(modelname, {})

        model_cache_dict = self.model_cache[modelname]
        for mid in modeldict:
            model = modeldict[mid]
            if mid in model_cache_dict:
                cached_model = model_cache_dict[mid]
                vars(cached_model).update(vars(model))
            else:
                model_cache_dict[mid] = model
        return [model_cache_dict[mid] for mid in modeldict]

    def _group_models_and_update_cache(self, models):
        grouped_by_type = {}
        for model in models:
            classname = model.__class__.__name__
            arr = grouped_by_type.setdefault(classname, [])
            if model is not None:
                arr.append(model)

        result = {}
        for clstype in grouped_by_type:
            result[clstype] = self._update_model_cache_helper(
                clstype, {m._primary_key: m for m in grouped_by_type[clstype]}
            )
        return result

    # TODO: support unsaved models as well
    def cached_find(self, model_class, id):
        if isinstance(id, list):
            return self.cached_where({"id": id}, model_class)
        cached_models = self.model_cache.get(model_class, {})
        found_model = cached_models.get(id, None)
        if found_model is None:
            found_model = self.interface(model_class).find(id)
        else:
            self.log.info(
                "CACHE found {} model with id={} in cache".format(model_class, id)
            )
        if found_model is None:
            return None
        return self._update_model_cache_helper(
            model_class, {found_model.id: found_model}
        )[0]

    def cached_where(self, query, model, primary_key="id"):
        if isinstance(query, str):
            server_models = self.interface(model).where(query)
            found_dict = {}
        elif [] in query.values():
            return []
        else:
            cached_models = self.model_cache.get(model, {})
            found, found_queries = self._find_matches(
                query, list(cached_models.values())
            )
            found_dict = {f.id: f for f in found}
            remaining_query = dict(query)
            if primary_key in query:
                found_ids = [q[primary_key] for q in found_queries]
                query_id_list = query[primary_key]
                if isinstance(query_id_list, str) or isinstance(query_id_list, int):
                    query_id_list = [query_id_list]
                remaining_ids = list(set(query_id_list).difference(set(found_ids)))
                remaining_query[primary_key] = remaining_ids
            self.log.info(
                "CACHE found {num} {model} models in cache using query {query}".format(
                    num=len(found_dict), model=model, query=self.log.pprint_data(query)
                )
            )

            # TODO: this code may be sketchy... here {'id': []}, really means we found
            #       all of the models..
            if primary_key in remaining_query and not remaining_query[primary_key]:
                return list(found_dict.values())
            server_models = self.interface(model).where(remaining_query)

        models_dict = OrderedDict({s.id: s for s in server_models})
        models_dict.update(found_dict)
        return self._update_model_cache_helper(model, models_dict)

    def _search_helper(self, pattern, filter_fxn, sample_type=None, **query):
        sample_type_id = None
        if sample_type is not None:
            if isinstance(sample_type, pydent_models.SampleType):
                sample_type_id = sample_type.id
            else:
                sample_type_id = self.find_by_name(sample_type, "SampleType").id

        if sample_type_id is not None:
            query.update({"sample_type_id": sample_type_id})

        model_list = self.list_models()
        self.log.info(
            "SEARCH found {} total models of type {}".format(
                len(model_list), self.model_name
            )
        )
        matches = filter_fxn(pattern, model_list)

        if not matches:
            return []

        if query:
            query.update({"id": matches})
            filtered = self.where(query, self.model_name)
        else:
            filtered = self.interface().find(matches)
        self.log.info(
            "SEARCH filtered to {} total models of type {}".format(
                len(filtered), self.model_name
            )
        )

        if self.use_cache:
            self.update_cache(filtered)
        return filtered

    def search(self, pattern, ignore_case=True, sample_type=None, **query):
        """Performs a regular expression search of Samples.

        :param pattern: regular expression pattern
        :type pattern: basestring
        :param ignore_case: whether to ignore case for regex search (default: True)
        :type ignore_case: bool
        :param sample_type: sample_type_name to filter samples (optional)
        :type sample_type: basestring
        :param query: additional query parameters to filter by
        :type query: dict
        :return: list of samples
        :rtype: list
        """

        def regex_filter(pattern, sample_list):
            matches = []
            for name in sample_list:
                args = []
                if ignore_case:
                    args.append(re.IGNORECASE)
                m = re.search(pattern, name, *args)
                if m is not None:
                    matches.append(name)
            return matches

        return self._search_helper(
            pattern, regex_filter, sample_type=sample_type, **query
        )

    def search_description(
        self, pattern, samples=None, sample_type=None, ignore_case=True
    ):
        """Search samples by their description.

        :param pattern: regex pattern
        :type pattern: basestring
        :param samples: samples to search. If left blank, a search to find samples will
            be performed
        :type samples: list
        :param sample_type: restrict to a particular sample type
        :type sample_type: name
        :param ignore_case: ignore case for regular pattern search (default True)
        :type ignore_case: bool
        :return: list of samples
        :rtype: list
        """
        matches = []
        args = []
        if samples is None:
            samples = self.search(".*", sample_type=sample_type)
        if ignore_case:
            args.append(re.IGNORECASE)
        for sample in samples:
            description = sample.description
            if description is None:
                continue
            if re.search(pattern, sample.description, *args):
                matches.append(sample)
        return matches

    def close_matches(self, pattern, sample_type=None, **query):
        """Finds samples whose names closely match the pattern.

        :param pattern: regular expression pattern
        :type pattern: basestring
        :param sample_type: sample_type_name to filter samples (optional)
        :type sample_type: basestring
        :param query: additional query parameters to filter by
        :type query: dict
        :return: list of samples
        :rtype: list
        """
        return self._search_helper(
            pattern, get_close_matches, sample_type=sample_type, **query
        )

    def list_field_values(self, model_ids, **query):
        """Lists sample field values.

        May supply an additional query to filter :class:`FieldValue`s.
        """
        query.update({"parent_class": self.model_name, "parent_id": model_ids})
        return self.where(query, "FieldValue")

    @staticmethod
    def _group_by_attribute(models, attribute):
        """Group models by the given attribute."""
        d = {}
        for s in models:
            arr = d.setdefault(getattr(s, attribute), [])
            arr.append(s)
        return d

    def new_sample(self, sample_type, name, description, project, properties=None):
        st = self.find_by_name(sample_type, "SampleType", primary_key="name")
        if st._get_deserialized_data().get("field_types", None) is None:
            fts = self.where(
                {"parent_class": "SampleType", "parent_id": st.id}, "FieldType"
            )
            st.field_types = fts
        return st.new_sample(name, description, project, properties=properties)

    # TODO: handle not-yet existant samples using rid
    def _retrieve_has_many_or_has_one(
        self, models, relationship_name, relation=None, strict=True
    ):
        """Performs exactly 1 query to fullfill some relationship for a list of
        models."""
        if not models:
            return []
        models = models[:]
        if relation is None:
            relation = models[0].get_relationships()[relationship_name]

        ref = relation.ref  # sample_id
        attr = relation.attr  # id
        model_class2 = relation.nested

        # todo: collect existing fullfilled relationship
        # todo: partition, then collect callback
        # todo: how to handle when model_attr is absent?, or just raise error?

        retrieve_query = relation.build_query(models)
        retrieved_models = self.where(retrieve_query, model_class2)
        self.log.info(
            "RETRIEVE retrieved {num} {cls} models using query {query}".format(
                num=len(retrieved_models),
                cls=model_class2,
                query=self.log.pprint_data(retrieve_query),
            )
        )

        if not retrieved_models:
            return []

        if relation.QUERY_TYPE == "query":
            model_dict = {m1.id: list() for m1 in models}
            for model in retrieved_models:
                model_ref = getattr(model, ref)
                if model_ref is not None:
                    if not strict:
                        model_dict.setdefault(model_ref, []).append(model)
                    else:
                        model_dict[model_ref].append(model)
                else:
                    self.log.error(
                        "RETRIEVE ref: {ref} {model_ref}, attr: {attr}".format(
                            ref=ref, attr=attr, model_ref=model_ref
                        )
                    )
        elif relation.QUERY_TYPE == "by_id":
            retrieved_dict = {getattr(m2, attr): m2 for m2 in retrieved_models}
            model_dict = {m1.id: None for m1 in models}
            missing_models = []
            for model in models:
                model_attr = getattr(model, attr)
                model_ref = getattr(model, ref)
                if model_ref is not None:
                    if model_attr is not None:
                        if model_ref not in retrieved_dict:
                            self.log.info(strict)
                        else:
                            model_dict[model_attr] = retrieved_dict[model_ref]
                    else:
                        self.log.error(
                            "attr: {attr}={model_attr}, ref: {ref}={model_ref}".format(
                                m1=model,
                                ref=ref,
                                attr=attr,
                                model_attr=model_attr,
                                model_ref=model_ref,
                            )
                        )
                    if model_ref not in retrieved_dict:
                        missing_models.append(model_ref)
            if missing_models:
                self.log.error(
                    "INCONSISTENT AQUARIUM DATABASE - There where {models} missing {cls} "
                    "models "
                    "from the Aquarium database, which were ignored by trident. "
                    "This happens when models are deleted from Aquarium which "
                    "results in an inconsistent server database. Trident was unable "
                    "resolve the following relationships which returned no models "
                    "from the server: "
                    "{cls}.where({attr}={missing})".format(
                        models=len(missing_models),
                        cls=model_class2,
                        missing=missing_models,
                        attr=attr,
                    )
                )
        else:
            raise BrowserException(
                "QUERY_TYPE '{}' for relation '{}' not recognized.".format(
                    relation.QUERY_TYPE, relationship_name
                )
            )
        for model in models:
            found_models = model_dict[getattr(model, attr)]
            setattr(model, relationship_name, found_models)

        return retrieved_models

    def _retrieve_has_many_through(
        self, models: List[ModelBase], relationship_name: str, strict: bool = True
    ):
        """Performs exactly 2 queries to establish a HasManyThrough
        relationship."""
        relation = models[0].get_relationships()[relationship_name]
        association_relation = models[0].get_relationships()[
            relation.through_model_attr
        ]
        attr = relation.attr
        ref = association_relation.ref

        # find other key
        association_class = ModelRegistry.get_model(association_relation.nested)
        association_relationships = association_class.get_relationships()
        other_ref = None
        for r in association_relationships:
            ar = association_relationships[r]
            if ar.nested == relation.nested:
                other_ref = r

        associations = self._retrieve_has_many_or_has_one(
            models, relation.through_model_attr, strict=strict
        )
        self._retrieve_has_many_or_has_one(associations, other_ref, strict=strict)

        associations_by_mid = {}
        for a in associations:
            associations_by_mid.setdefault(getattr(a, ref), []).append(a)

        all_models = []
        for m in models:
            mid = getattr(m, attr)
            if mid in associations_by_mid:
                associations = associations_by_mid[mid]
                _models = [getattr(a, other_ref) for a in associations]
                setattr(m, relationship_name, _models)
                all_models += _models
            else:
                setattr(m, relationship_name, None)
        return list(set(all_models))

    @staticmethod
    def _get_relation_from_model(model, relationship_name, strict):
        relationships = model.get_relationships()
        relation = relationships.get(relationship_name, None)
        if relation is None:
            if strict:
                msg = did_you_mean(relationship_name, list(relationships))
                if not msg:
                    msg = "Available relations: {}".format(pformat(list(relationships)))
                raise BrowserException(
                    "Relation '{}' not found in relationships for {}.\nHint: {}".format(
                        relationship_name, type(model), msg
                    )
                )
        return relation

    def retrieve(
        self,
        models: List[ModelBase],
        relationship_name: str,
        relation: BaseRelationship = None,
        strict: bool = True,
        force_refresh: bool = False,
    ) -> List[ModelBase]:
        """Retrieves a model relationship for the list of models. Compared to a
        `for` loop, `retrieve` is >10X faster for most queries.

        .. code-block:: python

            # FAST code example
            samples = browser.search(".*mCherry.*")
            items = browser.retrieve(samples, 'items')
            for s in samples:
                sample_items = s.items
                # do something with items


        .. code-block:: python

            # very SLOW code example. DO NOT USE.
            samples = browser.search(".*mCherry.*")
            for s in samples:
                sample_items = s.items
                # do something with itmes


        :param models: list of models to retrieve the attribute
        :type models: list
        :param relationship_name: name of the attribute to retrieve
        :type relationship_name: basestring
        :param relation: the relation to retrieve (operational)
        :type relation: pydent.relationships.Relation
        :param strict: wither to ignore database inconsistencies
        :type bool
        :return: list of models retrieved
        :rtype: list
        """
        if not models:
            return []
        self.log.info('RETRIEVE retrieving "{}"'.format(relationship_name))
        model_classes = {m.__class__.__name__ for m in models}
        assert (
            len(model_classes) == 1
        ), "Models must be all of the same BaseModel, but found {}".format(
            model_classes
        )
        if relation is None:
            relation = self._get_relation_from_model(
                models[0], relationship_name, strict
            )
            if relation is None:
                return []
        else:
            if relationship_name in models[0].get_relationships():
                raise BrowserException(
                    'Cannot add new relationship "{}" because it already exists in the '
                    "model definition".format(relationship_name)
                )
        # TODO: add relationship handler for Many and One
        if relation.__class__.__name__ not in self.ACCEPTED_GET_RELATION_TYPES:
            raise BrowserException(
                'retrieve is not supported for the "{}" relationship'.format(
                    relation.__class__.__name__
                )
            )
        self.log.info("RETRIEVE {}: {}".format(relationship_name, relation))

        if not force_refresh:
            needs_refresh = [
                m for m in models if not m.is_deserialized(relationship_name)
            ]
            no_refresh = [m for m in models if m.is_deserialized(relationship_name)]
        else:
            needs_refresh = models
            no_refresh = []

        if needs_refresh:
            if hasattr(relation, "through_model_attr"):
                found_models = self._retrieve_has_many_through(
                    needs_refresh, relationship_name, strict=strict
                )
            else:
                found_models = self._retrieve_has_many_or_has_one(
                    needs_refresh, relationship_name, relation, strict=strict
                )
        else:
            found_models = []

        self.log.info(
            'RETRIEVE retrieved {} for "{}"'.format(
                len(found_models), relationship_name
            )
        )
        for model in no_refresh:
            val = getattr(model, relationship_name)
            if isinstance(val, list):
                found_models += val
            else:
                found_models.append(val)
        return list(set(found_models))

    def recursive_retrieve(
        self,
        models: List[ModelBase],
        relations: Union[str, List[BaseRelationship], Dict],
        strict: bool = True,
        force_refresh: bool = False,
    ):
        """Efficiently retrieve a model relationship recursively from an
        iterable. The relations_dict iterable may be either a list or a
        dictionary. For example, the following will collect all of the
        field_values and their incoming and outgoing wires, the connecting
        field_values, and finally those FieldValues' operations.

        .. code-block::

            relation_dict = {
                "field_values": {
                    "wires_as_dest": {
                        "source": "operation",
                        "destination": "operation"
                    },
                    "wires_as_source": {
                        "source": "operation",
                        "destination": "operation"
                    },
                }
            }
            browser.retrieve(operations, relation_dict)


        :param models: models to retrieve from
        :type models: list
        :param relations: the relation to retrieve. This may either be a string (by
            attribute name), a list, or a dict.
        :type relations: list|dict|basestring
        :param strict: wither to ignore database inconsistencies
        :param force_refresh:
        :type bool
        :return: dictionary of all models retrieved grouped by the attribute name that
            retrieved them.
        :rtype: dictionary
        """
        self.log.info("RETRIEVE recursively retrieving {}".format(relations))
        if isinstance(relations, str):
            self.log.info('RETRIEVE retrieving "{}"'.format(relations))
            return {
                relations: self.retrieve(
                    models, relations, strict=strict, force_refresh=force_refresh
                )
            }
        elif (
            isinstance(relations, list)
            or isinstance(relations, set)
            or isinstance(relations, dict)
            or isinstance(relations, tuple)
        ):
            models_by_attr = {}
            for relation_name in relations:
                models_by_attr.setdefault(relation_name, [])
                new_models = self.retrieve(
                    models, relation_name, strict=strict, force_refresh=force_refresh
                )
                models_by_attr[relation_name] += new_models
                if isinstance(relations, dict):
                    _models_by_attr = self.recursive_retrieve(
                        new_models,
                        dict(relations).pop(relation_name),
                        strict=strict,
                        force_refresh=force_refresh,
                    )
                    for attr in _models_by_attr:
                        _models = _models_by_attr[attr]
                        if attr in models_by_attr:
                            models_by_attr[attr] += _models
                        else:
                            models_by_attr[attr] = _models_by_attr[attr]
            return models_by_attr
        elif not strict:
            return []
        else:
            raise BrowserException(
                "Type {} for is not recognized for recursive_retrieve".format(
                    type(relations)
                )
            )

    @classmethod
    def sample_network(cls, samples, reverse=False, g=None):
        """Build a DAG of :class:`Samples <pydent.models.Sample>` from their.

        :class:`FieldValues <pydent.models.FieldValue>`.

        .. versionadded:: 0.1.5a7
            method added

        :param samples: list of samples
        :param reverse: whether to reverse the edges of the final graph
        :param g: the graph
        :return:
        """

        def get_models(model):
            for fv in model.field_values:
                if fv.sample and issubclass(type(fv.sample), ModelBase):
                    yield fv.sample, {"field_value_name": fv.name}

        def cache_func(models):
            sess = models[0].session
            sess.browser.get(models, {"field_values": "sample"})

        return cls.relationship_network(
            samples, get_models=get_models, cache_func=cache_func, reverse=reverse, g=g
        )

    @classmethod
    def relationship_network(
        cls,
        models: List[ModelBase],
        get_models: Callable,
        cache_func: Callable = None,
        key_func: Callable = None,
        reverse: bool = False,
        g: nx.DiGraph = None,
        strict_cache: bool = True,
    ):
        """Build a DAG of related models based on some relationships. By
        default are built from a model using (model.__class__.__name__,
        model._primary_key)

        .. versionadded:: 0.1.5a7
            method added

        .. seealso::
            Usage example :meth:`sample_network <pydent.browser.Browser.sample_network>`

        :param models: list of models
        :param get_models: A function that takes in a single instance of a Model
            and returns an iterable of Tuples of (model, data),
            which is used to build edges.
        :param key_func: An optional function that takes in a single instance of
            a Model and returns a key and some node data
        :param cache_func: A function that takes in a list of models and caches
            some results.
        :param g: optional nx.DiGraph
        :param reverse: whether to reverse the edge list
        :param strict_cache: if True, if a request occurs after the cache step,
            a ForbiddenRequestException will be raised.
        :return: the relationship graph
        """

        if key_func is None:

            def key_func(m):
                return (m.__class__.__name__, m._primary_key), {}

        if g is None:
            g = nx.DiGraph()

        models = list(models)

        if not models:
            return g

        for m in models:
            key, ndata = key_func(m)
            g.add_node(key, attr_dict=ndata)

        visited = list(g.nodes)

        if cache_func:
            cache_func(models)

        new_models = []
        new_edges = []

        if strict_cache:
            kwargs = {"using_requests": False, "session_swap": True}
        else:
            kwargs = {}

        with models[0].session(**kwargs):
            try:
                for m1 in models:
                    for m2, data in get_models(m1):
                        if key_func(m2)[0] not in visited:
                            new_models.append(m2)
                        new_edges.append((key_func(m2)[0], key_func(m1)[0], data))
            except ForbiddenRequestError as e:
                msg = (
                    "An exception occurred during {f} while strict_cache == True.\n"
                    "This is most likely due to the cache_func not being thorough.\n"
                    "{}".format(str(e))
                )
                raise e.__class__(msg)

        for n1, n2, edata in new_edges:
            if reverse:
                g.add_edge(n2, n1, attr_dict=edata)
            else:
                g.add_edge(n1, n2, attr_dict=edata)

        return cls.relationship_network(
            new_models,
            get_models,
            cache_func,
            g=g,
            reverse=reverse,
            strict_cache=strict_cache,
        )

    # def relationship_network(self, models, get_models, cache_func, g=None):
    #     """
    #
    #
    #     .. versionadded:: 0.1.5a7
    #
    #
    #     :param models:
    #     :param get_models:
    #     :param g:
    #     :return:
    #     """
    #
    #     def model_to_key(m):
    #         return (m.__class__.__name__, m.id)
    #
    #     if g is None:
    #         g = nx.DiGraph()
    #     for m in models:
    #         g.add_node(model_to_key(m))
    #
    #     cache_func(models)
    #
    #     for m in models:
    #         next_models = get_models(m)

    def get(
        self,
        models: List[ModelBase],
        relations: List[BaseRelationship] = None,
        query: dict = None,
        strict: bool = True,
        force_refresh: bool = False,
    ) -> Union[Dict[str, List[ModelBase]], List[ModelBase]]:
        """

        :param models:
        :param relations:
        :param query:
        :param strict:
        :param force_refresh:
        :return:
        """
        if isinstance(models, ModelBase):
            models = [models]
        elif isinstance(models, str):
            models = list(self.model_cache.get(models, {}).values())
            if query:
                models, _ = self._find_matches(query, models)
        if relations:
            if isinstance(relations, str):
                return self.retrieve(
                    models, relations, strict=strict, force_refresh=force_refresh
                )
            else:
                return self.recursive_retrieve(
                    models, relations, strict=strict, force_refresh=force_refresh
                )
        else:
            return models

    def export_samples_to_csv(self, samples, out):
        """Exports the samples to a csv (for Aquarium import)

        :param samples: list of samples
        :type samples: list
        :param out: output path of csv file
        :type out: basestring
        :return: pandas dataframe used for the csv
        :rtype: pandas.DataFrame
        """
        df = self.samples_to_df(samples)
        df.to_csv(out)
        return df

    def samples_to_rows(self, samples, sample_resolver=None):
        """Return row of dictionaries containing sample information and their
        properties. Can be imported into a pandas DataFrame:

        .. code-block:: python

            import pandas

            df = pandas.DataFrame(samples_to_rows(samples))


        :param samples: samples, all of same sample type
        :type samples: list
        :param sample_resolver: callable to resolve a sample object if the field value
            property contains a sample.
        Defaults to return the sample anme
        :type sample_resolver: callable
        :return: list of dictionaries containing sample info and properties
        :rtype: list
        """

        assert (
            len({s.sample_type_id for s in samples}) == 1
        ), "Samples be the of the same SampleType"
        sample_type = samples[0].sample_type

        # self.recursive_retrieve(samples, {
        #     "field_values": "sample",
        #     "sample_type": {
        #         "field_types"
        #     }
        # })
        if sample_resolver is None:

            def default_resolver(sample):
                if isinstance(sample, pydent_models.Sample):
                    return sample.name
                if isinstance(sample, list):
                    resolved_samples = []
                    for s in sample:
                        if isinstance(s, pydent_models.Sample):
                            resolved_samples.append(s.name)
                        else:
                            resolved_samples.append(s)
                    sample = resolved_samples
                return sample

            sample_resolver = default_resolver
        rows = []
        for s in samples:
            row = {
                sample_type.name: s.name,
                "Description": s.description,
                "Project": s.project,
            }
            props = {k: sample_resolver(v) for k, v in s.properties.items()}
            row.update(props)
            rows.append(row)
        return rows

    # TODO: add to change log
    @staticmethod
    def _inspect_header(model, name=None, _id="id"):
        if name is not None:
            name = eval("model.{}".format(name))
        else:
            name = ""
        return "{model_class}: {_id} {name}".format(
            model_class=model.__class__.__name__,
            name=name,
            _id=eval("model.{}".format(_id)),
        )

    # TODO: add to change log
    @staticmethod
    def _inspect_rows(model, attributes, indent=2):
        rows = []
        for attr in attributes:
            rows.append(
                "{attr}: {val}".format(attr=attr, val=eval("model.{}".format(attr)))
            )
        sep = "\n" + " " * indent
        s = sep + sep.join(rows)
        return s

    # TODO: add to change log
    @classmethod
    def inspect(cls, model, header, attributes):
        model_header = cls._inspect_header(model, **header)
        rows = cls._inspect_rows(model, attributes)
        return "{header}{rows}".format(header=model_header, rows=rows)



"""
Exceptions (:mod:`pydent.exceptions`)
=====================================

.. currentmodule:: pydent.exceptions

Trident exceptions
"""


class TridentBaseException(Exception):
    """Base exception for all trident errors."""


class TridentRequestError(TridentBaseException):
    """There was an ambiguous exception that occurred handling your request."""

    def __init__(self, msg, response):
        super().__init__(msg)
        self.response = response


class AquariumError(TridentBaseException):
    """Aquarium raised an error."""


class AquariumModelNotFound(TridentBaseException):
    """Returned when Aquarium could not find a given model."""


class ForbiddenRequestError(TridentBaseException):
    """Raised when Trident attempts to make a request after requests have been
    explicitly turned off."""


class TridentJSONDataIncomplete(TridentBaseException):
    """JSON data contains a null value and may be incomplete."""


class TridentLoginError(TridentBaseException):
    """Trident is not properly connected to the server.

    Verify login credentials are correct.
    """


class TridentTimeoutError(TridentBaseException):
    """Trident took too long to respond."""


class TridentModelNotFoundError(TridentBaseException):
    """Trident could not find model in list of models."""


class AquariumModelError(TridentBaseException):
    """An error occurred with this Aquarium model."""


class NoSessionError(TridentBaseException):
    """There was no session attached to the model, but one is required."""


class SessionAlreadySet(TridentBaseException):
    """Cannot set session to models with a session already set."""


class PlannerException(Exception):
    """Generic planner Exception."""


class PlannerVerificationException(Exception):
    """Raised when object is not in plan but is required."""



"""pydent."""
# without nest_asyncio, Jupyter notebooks will crash if an async method is used
import getpass

import nest_asyncio

from .__version__ import __authors__
from .__version__ import __homepage__
from .__version__ import __repo__
from .__version__ import __title__
from .__version__ import __version__
from .aqsession import AqSession
from .base import ModelBase
from .base import ModelRegistry
from .browser import Browser
from .inventory_updater import save_inventory
from .planner import Planner
from .utils import logger
from .utils import pprint

nest_asyncio.apply()


def login(user, url, password=None):
    """Create a new AqSession instance from login information.

    :param user: user login
    :type user: basestring
    :param url: server url
    :type url: basestring
    :param password: optional password. If left blank, a secure prompt will ask for the password.
    :type password: basestring | None
    :return: the session
    :rtype: AqSession
    """
    if password is None:
        return AqSession(user, getpass.getpass(), url)
    return AqSession(user, password, url)



"""
Serialization (:mod:`pydent.marshaller`)
========================================================

.. versionadded 0.1
    This module replaces the much slower Marshmallow library. Speed improvements >10X.

.. currentmodule:: pydent.marshaller

This module provides speedy serialization and deserialization
functionality for Trident models. Models that inherit the
:class:`SchemaModel <pydent.marshaller.base.SchemaModel>` will
become registered with their own serialization/deserialization
schema. To register the schema,
use the :meth:`add_schema <pydent.marshaller.base.add_schema>`
decorator on the class. The
:meth:`dump <pydent.marshaller.base.SchemaModel.dump>` method will serialize
and object while the :meth:`dump <pydent.marshaller.base.SchemaModel.load>` method
will deserialize JSON data to an object.

.. note::
    This is only relevant to Trident **developers**, rather than regular users.

.. autosummary::
    :toctree: generated/

    SchemaModel
    ModelRegistry
    SchemaRegistry
    descriptors
    exceptions
    utils

Fields
------

.. currentmodule:: pydent.marshaller.fields

Fields are descriptors that control how attributes and data is serialized/deserialized.

.. autosummary::
    :toctree: generated/

    Alias
    Callback
    Field
    FieldABC
    Nested
    Relationship
"""
from pydent.marshaller import descriptors
from pydent.marshaller import exceptions
from pydent.marshaller import fields
from pydent.marshaller.base import add_schema
from pydent.marshaller.base import SchemaModel
from pydent.marshaller.registry import ModelRegistry
from pydent.marshaller.registry import SchemaRegistry



"""PlannerLayout."""
from collections import OrderedDict
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import networkx as nx

from pydent.models import Operation
from pydent.models import Plan
from pydent.planner.utils import _id_getter
from pydent.planner.utils import get_subgraphs
from pydent.utils import make_async

# custom types
NodeType = Union[str, int]
PlannerGraphType = "PlannerGraph"
PointValue = float
PointType = Tuple[PointValue, PointValue]


class PlannerGraph:
    """Representation of an Aquarium plan as a directed graph.

    Positions :class:`Operations <pydent.models.Operation>` for the
    Aquarium GUI
    """

    def __init__(self, nxgraph: nx.DiGraph = None):
        """Initializes a Plan graph.

        :param nxgraph:
        """
        if nxgraph is None:
            nxgraph = nx.DiGraph()
        self.nxgraph = nxgraph
        self.annotations = []

    @classmethod
    def from_plan(cls, plan: Plan) -> PlannerGraphType:
        """Creates a graph from a :class:`pydent.models.Plan` instance."""
        G = nx.DiGraph()
        planner_graph = cls(G)

        # store reference of operation
        fv_to_op_dict = {}
        if not plan.operations:
            return planner_graph
        for op in plan.operations:
            for fv in op.field_values:
                fv_to_op_dict[fv.rid] = op

        @make_async(10, progress_bar=False)
        def add_wires(wires):
            if not wires:
                return
            missing_operations = []
            for wire in wires:
                # TODO:
                try:
                    from_id = _id_getter(fv_to_op_dict[wire.source.rid])
                    to_id = _id_getter(fv_to_op_dict[wire.destination.rid])
                except KeyError as e:
                    pass
                    raise e
                if from_id is not None and to_id is not None:
                    if from_id not in G:
                        missing_operations.append(from_id)
                    if to_id not in G:
                        missing_operations.append(to_id)
                    if from_id in G and to_id in G:
                        G.add_edge(from_id, to_id, wire=wire)
            return wires

        @make_async(10, progress_bar=False)
        def add_ops(ops: Iterable[Operation]):
            if not ops:
                return
            for op in ops:
                planner_graph._add_operation(op)
            return ops

        add_ops(plan.operations)
        add_wires(plan.wires)

        # fix operation coordinates if None
        for op in planner_graph.operations:
            if op.x is None:
                op.x = 0
            if op.y is None:
                op.y = 0

        return planner_graph

    @property
    def nodes(self):
        return self.nxgraph.nodes

    @property
    def operations(self) -> Generator[Operation, None, None]:
        for node in self.nxgraph.nodes:
            if "operation" in self.nxgraph.nodes[node]:
                yield self.nxgraph.nodes[node]["operation"]

    def iter_operations(
        self
    ) -> Generator[Tuple[Union[str, int], Operation], None, None]:
        for node in self.nxgraph.nodes:
            if "operation" in self.nxgraph.nodes[node]:
                yield (node, self.nxgraph.nodes[node]["operation"])

    def _add_operation(self, op: Operation):
        """Adds an Operation to the Layout.

        :param op: operation
        :type op: pydent.models.Operation
        :return: None
        :rtype: None
        """
        return self.nxgraph.add_node(_id_getter(op), operation=op)

    def subgraph(self, nodes: List[NodeType]) -> Optional[Type["PlannerGraph"]]:
        """Returns a subgraph layout from a list of node_ids."""
        return self.__class__(nxgraph=self.nxgraph.subgraph(nodes))

    def nodes_to_ops(self, nodes: List[NodeType]):
        """Returns operations from a list of node_ids."""
        return [self.nxgraph.nodes[n]["operation"] for n in nodes]

    def ops_to_nodes(self, ops: List[Operation]):
        """Returns node_ids for each operation."""
        return [_id_getter(op) for op in ops]

    def ops_to_subgraph(self, ops: List[Operation]) -> PlannerGraphType:
        """Returns a sub-layout containing only the operations."""
        return self.subgraph(self.ops_to_nodes(ops))

    def predecessors(self, node: NodeType):
        """Return predecessor nodes.

        :param node: node id
        :type node: basestring
        :return: list of node ids
        :rtype: list
        """
        return self.nxgraph.predecessors(node)

    def successors(self, node: NodeType) -> Generator[NodeType, None, None]:
        """Return successors nodes.

        :param node: node id
        :type node: basestring
        :return: list of node ids
        :rtype: list
        """
        return self.nxgraph.successors(node)

    def predecessor_subgraph(self, subgraph: PlannerGraphType) -> PlannerGraphType:
        n_bunch = self.collect_predecessors(subgraph.roots())
        return self.subgraph(n_bunch)

    def successor_subgraph(self, layout):
        n_bunch = self.collect_successors(layout.leaves())
        return self.subgraph(n_bunch)

    def collect_predecessors(self, nodes: List[NodeType]) -> List[NodeType]:
        """Return all predecessors from a list of nodes.

        :param nodes: list of node ids
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        predecessors = []
        for node in nodes:
            predecessors += self.predecessors(node)
        return list(set(predecessors))

    def collect_successors(self, nodes: List[NodeType]) -> List[NodeType]:
        """Return all successors from a list of nodes.

        :param nodes: list of node ids
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        successors = []
        for node in nodes:
            successors += self.successors(node)
        return list(set(successors))

    def ops_to_predecessors(self, ops: List[Operation]) -> List[Operation]:
        """Return all predecessor operations from a list of operations.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: list of node ids
        :rtype: list
        """
        nodes = self.collect_predecessors(self.ops_to_nodes(ops))
        return self.nodes_to_ops(nodes)

    def ops_to_successors(self, ops: List[Operation]) -> List[Operation]:
        """Return all successor operations from a list of operations.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: list of node ids
        :rtype: list
        """
        nodes = self.collect_successors(self.ops_to_nodes(ops))
        return self.nodes_to_ops(nodes)

    def leaves(self):
        """Returns the leaves of this graph."""
        leaves = []
        for n in self.nodes:
            if len(list(self.successors(n))) == 0:
                leaves.append(n)
        return leaves

    def roots(self):
        """Returns the roots of this graph."""
        roots = []
        for n in self.nodes:
            if len(list(self.predecessors(n))) == 0:
                roots.append(n)
        return roots

    def ops_to_leaves(self):
        return self.nodes_to_ops(self.leaves())

    def ops_to_roots(self):
        return self.nodes_to_ops(self.roots())

    def __len__(self):
        return len(self.nxgraph)


class PlannerLayout(PlannerGraph):
    """Layout module for :class:`Planner <pydent.planner.Planner>`.

    Positions :class:`Operations <pydent.models.Operation>` for the
    Aquarium GUI
    """

    TOP_RIGHT = (100, 100)
    BOX_DELTA_X = 170
    BOX_DELTA_Y = 70
    BOX_WIDTH = 150
    BOX_HEIGHT = 70
    STATUS_COLORS = {
        "waiting": "orange",
        "error": "red",
        "pending": "yellow",
        "running": "green",
        "delayed": "magenta",
        "done": "black",
        "planning": "grey",
    }

    def get_independent_graphs(self):
        """Finds all independent subgraphs.

        :return: list of PlannerLayout
        :rtype: list
        """
        subgraphs = get_subgraphs(self.nxgraph)
        sublayouts = [self.__class__(nxgraph=g) for g in subgraphs]
        return sublayouts

    def _topological_sort(self):
        subgraphs = self.get_independent_graphs()
        for subgraph in subgraphs:
            subgraph._topological_sort_helper()
        self.arrange_layouts(subgraphs)

    def topo_sort_in_place(self):
        cx, cy = self.midpoint()
        self._topological_sort()
        cx2, cy2 = self.midpoint()
        self.translate(cx - cx2, cy - cy2)

    def topo_sort(self):
        """Does a topological sort of the operations in this layout.

        Discovers individual subgraphs and topologically sorts each subgraph,
        and then aligns subgraphs horizontally.

        :return: None
        :rtype: None
        """
        self._topological_sort()
        self.move(*self.TOP_RIGHT)

    def prettify(self):
        return self.topo_sort()

    @classmethod
    def arrange_layouts(cls, layouts):
        x = 0
        y = 0
        for layout in layouts:
            layout.move(x, y)
            x += layout.width + cls.BOX_DELTA_X
            y = layout.y

    def to_grid(
        self, columns: int, axis: int = 1, border_x: int = None, border_y: int = None
    ):
        """Arrange layouts in a grid format.

        :param columns: maximum number of columns (or rows when axis=0)
        :type columns: int
        :param axis: which axis to limit. Default: 1 (columns) or 0 (rows)
        :type axis: int
        :param border_x: (optional) separation between each cell in the grid
        :type border_x: int
        :param border_y: (optional) separation between each cell in the grid
        :type border_y: int
        :return: layout
        :rtype: layout
        """
        layouts = self.get_independent_graphs()
        if len(layouts) == 1:
            self.move(100, 100)
            return self

        # make grid assignments
        assignments = []
        c = 0
        r = 0
        for layout in layouts:
            assignments.append((r, c, layout))
            c += 1
            if c >= columns:
                r += 1
                c = 0

        # find cell height and width
        if border_x is None:
            border_x = self.BOX_DELTA_X * 2
        if border_y is None:
            border_y = self.BOX_DELTA_Y * 2
        grid_cell_height = 0
        grid_cell_width = 0
        for layout in layouts:
            if layout.height > grid_cell_height:
                grid_cell_height = layout.height
            if layout.width > grid_cell_width:
                grid_cell_width = layout.width
        grid_cell_height += border_y
        grid_cell_width += border_x

        for r, c, layout in assignments:
            if axis == 0:
                r, c = c, r
            layout.center(c * grid_cell_width, r * grid_cell_height)
        self.move(100, 100)
        return self

    def align_x_midpoints_to(self, other_layout):
        """Align the midpoint x-coordinate of this layout with that of another.

        :param other_layout: the other canvas layout
        :type other_layout: PlannerLayout
        :return: None
        :rtype: None
        """
        if len(other_layout.nxgraph) == 0:
            return
        other_x, _ = other_layout.midpoint()
        x, _ = self.midpoint()
        deltax = other_x - x
        self.translate(deltax, 0)

    def align_y_midpoints_to(self, other_layout):
        """Align this midpoint y-coordinates of this layout with another
        layout.

        :param other_layout: the other canvas layout
        :type other_layout: PlannerLayout
        :return: None
        :rtype: None
        """
        if len(other_layout.nxgraph) == 0:
            return
        _, other_y = other_layout.midpoint()
        self.y = other_y

    def align_x_of_ops(self, ops1: List[Operation], ops2: List[Operation]):
        """Aligns a set of Operations with another set of Operations within
        this PlannerLayout.

        :param ops1: list of :class:`pydent.models.Operation`
        :type ops1: list
        :param ops2: list of :class:`pydent.models.Operation`
        :type ops2: list
        :return: None
        :rtype: None
        """
        layout1 = self.ops_to_subgraph(ops1)
        layout2 = self.ops_to_subgraph(ops2)
        return layout1.align_x_midpoints_to(layout2)

    @property
    def xy(self) -> PointType:
        return self.bounds()[0]

    @property
    def x(self) -> PointValue:
        return self.xy[0]

    @property
    def y(self) -> PointValue:
        return self.xy[1]

    @x.setter
    def x(self, new_x: PointValue):
        deltax = new_x - self.x
        self.translate(deltax, 0)

    @y.setter
    def y(self, new_y: PointValue):
        delta_y = new_y - self.y
        self.translate(0, delta_y)

    def bounds(self) -> Tuple[PointType, PointType]:
        """Returns the bounding box of the layout as the upper-left and lower-
        right coordinates.

        :return: x,y coordinate
        :rtype: tuple
        """
        x_arr = [op.x for op in self.operations]
        y_arr = [op.y for op in self.operations]
        return (min(x_arr), min(y_arr)), (max(x_arr), max(y_arr))

    def midpoint(self) -> PointType:
        """Returns the midpoint x,y coordinates of the layout.

        :return: x,y coordinate
        :rtype: tuple
        """
        ul, lr = self.bounds()
        x, y = (lr[0] - ul[0]) / 2 + ul[0], (lr[1] - ul[1]) / 2 + ul[1]
        return x, y

    def translate(
        self, delta_x: PointValue, delta_y: PointValue
    ) -> Generator[Operation, None, None]:
        """Translates the layout by the x,y coordinates.

        :param delta_x: delta x
        :type delta_x: int
        :param delta_y: delta y
        :type delta_y: int
        :return: None
        :rtype: None
        """
        for op in self.operations:
            op.x += delta_x
            op.y += delta_y
        return self.operations

    def move(self, x: PointValue, y: PointValue):
        self.x = x
        self.y = y

    def center(self, x: PointValue, y: PointValue):
        self.move(x - self.width / 2, y - self.height / 2)

    @property
    def width(self):
        return self.bounds()[1][0] - self.bounds()[0][0]

    @property
    def height(self) -> PointValue:
        return self.bounds()[1][1] - self.bounds()[0][1]

    def pos(self) -> Dict[NodeType, PointType]:
        pos = {}
        for n, op in self.iter_operations():
            pos[n] = (op.x, -op.y)
        return pos

    def _status_colors(self) -> List[str]:
        node_color = [
            self.STATUS_COLORS.get(op.status, "rosybrown") for op in self.operations
        ]
        return node_color

    # TODO: minimize crossings
    def _topological_sort_helper(self):
        """Attempt a rudimentary topological sort on the plan."""

        _x, _y = self.TOP_RIGHT

        y = _y
        delta_x = self.BOX_DELTA_X
        delta_y = -self.BOX_DELTA_Y

        max_depth = {}
        roots = self.roots()
        for root in roots:
            depths = nx.single_source_shortest_path_length(self.nxgraph, root)
            for n, d in depths.items():
                max_depth[n] = max(max_depth.get(n, d), d)

        # push roots 'up' so they are not stuck on layer one
        for root in roots:
            successors = list(self.successors(root))
            if len(successors) > 0:
                min_depth = min([max_depth[s] for s in successors])
                max_depth[root] = min_depth - 1

        by_depth = OrderedDict()

        for node, depth in max_depth.items():
            by_depth.setdefault(depth, [])
            by_depth[depth].append(node)
        for depth in sorted(by_depth):
            op_ids = by_depth[depth]

            # sort by predecessor_layout
            predecessor_avg_x = []
            x = 0
            for op_id in op_ids:
                predecessors = list(self.predecessors(op_id))
                if len(predecessors) > 0:
                    x, _ = self.subgraph(predecessors).midpoint()
                    predecessor_avg_x.append((x, op_id))
                else:
                    predecessor_avg_x.append((x + 1, op_id))
            sorted_op_ids = [
                op_id for _, op_id in sorted(predecessor_avg_x, key=lambda x: x[0])
            ]

            x = _x
            # sorted_op_ids = sorted(op_ids)
            ops = self.nodes_to_ops(sorted_op_ids)
            for op in ops:
                op.x = x
                op.y = y
                x += delta_x
            layer = self.subgraph(op_ids)
            predecessor_layout = self.predecessor_subgraph(layer)
            layer.align_x_midpoints_to(predecessor_layout)
            y += delta_y

        # readjust
        self.move(_x, _y)

    def align_ops_with_successors(self, ops: List[Operation]):
        """Aligns the Operations to their successors by midpoints.

        Used in topological sorting of plans.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: None
        :rtype: None
        """
        successors = self.ops_to_successors(ops)
        self.align_x_of_ops(ops, successors)

    def align_ops_with_predecessors(self, ops: List[Operation]):
        """Aligns the Operations to their predecessors by midpoints.

        Used in topological sorting of plans.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: None
        :rtype: None
        """
        predecessors = self.ops_to_predecessors(ops)
        self.align_x_of_ops(ops, predecessors)

    def draw(self):
        return nx.draw(self.nxgraph, pos=self.pos(), node_color=self._status_colors())



"""Models related to protocol code, which is associated with operation
types."""
from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.relationships import HasManyGeneric
from pydent.relationships import HasOne
from pydent.relationships import HasOneFromMany
from pydent.relationships import One


@add_schema
class Code(ModelBase):
    """A Code model."""

    fields = dict(
        user=HasOne("User"),
        operation_type=One("OperationType", callback="get_parent", callback_args=None),
        library=One("Library", callback="get_parent", callback_args=None),
    )

    def get_parent(self, parent_class, *args):
        if parent_class != self.parent_class:
            return None
        return self.session.model_interface(self.parent_class).find(self.parent_id)

    def update(self):
        # since they may not always be tied to specific parent
        # controllers
        self.session.utils.update_code(self)


@add_schema
class Library(ModelBase):
    """A Library model."""

    fields = dict(
        codes=HasManyGeneric("Code"),
        source=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "Library", "name": "source"},
        ),
    )

    def code(self, accessor):
        """Reminant from previous API."""
        # raise DeprecationWarning("This method is depreciated. Use '.source' directly")
        if accessor == "source":
            return self.source
        return None



"""
Planner (:mod:`pydent.planner`)
===============================

.. versionadded:: 0.1
    Planning module created

.. currentmodule:: pydent.planner

The planner module allows users to plan and submit simple and complex scientific
experiments and upload the plan to the Aquarium server for execution.

.. seealso::
    For more information on how to plan experiments using Trident,
    see :ref:`Planning Docs <planning>` and the :ref:`Advanced Topics <advancedtopics>`.

Classes
-------

.. autosummary::
    :toctree: generated/

    Planner
    PlannerLayout

Exceptions
----------

.. autosummary::
    :toctree: generated/

    PlannerException
"""
from pydent.planner.graph import PlannerLayout
from pydent.planner.planner import Planner
from pydent.planner.planner import PlannerException



"""Models related to samples, like Samples and SampleTypes."""
import json
from collections.abc import Sequence
from warnings import warn

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import JSONSaveMixin
from pydent.models.field_value_mixins import FieldTypeInterface
from pydent.models.field_value_mixins import FieldValueInterface
from pydent.relationships import HasMany
from pydent.relationships import HasOne


@add_schema
class Sample(FieldValueInterface, ModelBase):
    """A Sample model."""

    fields = dict(
        # sample relationships
        sample_type=HasOne("SampleType"),
        items=HasMany("Item", ref="sample_id"),
        field_values=HasMany(
            "FieldValue", ref="parent_id", additional_args={"parent_class": "Sample"}
        ),
    )

    METATYPE = "sample_type"

    def __init__(
        self,
        name=None,
        project=None,
        description=None,
        sample_type=None,
        sample_type_id=None,
        properties=None,
        field_values=None,
    ):
        """

        :param name:
        :type name:
        :param project:
        :type project:
        :param description:
        :type description:
        :param sample_type_id:
        :type sample_type_id:
        :param properties:
        :type properties:
        """
        super().__init__(
            name=name,
            project=project,
            description=description,
            sample_type_id=sample_type_id,
            sample_type=sample_type,
            field_values=field_values,
            items=None,
        )

        if properties is not None:
            self.update_properties(properties)

    @property
    def identifier(self):
        """Return the identifier used by Aquarium in autocompletes."""
        return "{}: {}".format(self.id, self.name)

    def field_value(self, name):
        """Returns the :class:`FieldValue` associated with the sample by its
        name. If the there is more than one FieldValue with the same name (as
        in field_value arrays), it will return the first FieldValue. See the.

        :meth:`Sample.field_value_array` method.

        :param name: name of the field value
        :type name: str
        :return: the field value
        :rtype: FieldValue
        """
        return self.get_field_value(name, None)

    def field_value_array(self, name):
        return self.get_field_value_array(name, None)

    def _property_accessor(self, fv):
        ft = self.safe_get_field_type(fv)
        if ft:
            if ft.ftype == "sample":
                return fv.sample
            else:
                if ft.ftype == "number":
                    if isinstance(fv.value, str):
                        return json.loads(fv.value)
                return fv.value

    @property
    def properties(self):
        return self._field_value_dictionary(lambda ft: ft.name, self._property_accessor)

    def update_properties(self, prop_dict):
        """Update the FieldValues properties for this sample.

        :param prop_dict: values to update
        :type pro fp_dict: dict
        :return: self
        :rtype: Sample
        """
        ft_dict = {ft.name: ft for ft in self.get_field_types()}
        for name, val in prop_dict.items():
            ft = ft_dict[name]
            if ft.is_parameter():
                key = "value"
            else:
                key = "sample"
            if issubclass(type(val), Sequence) and ft.array:
                self.set_field_value_array(name, None, [{key: v} for v in val])
            else:
                self.set_field_value(name, None, {key: val})

    def create(self):
        return self.session.utils.create_samples([self])

    def save(self):
        if self.id:
            self.update()
        else:
            self.create()

    def update(self):
        for fv in self.field_values:
            fv.reload(fv.save())

        new_fvs = self.field_values
        server_fvs = self.session.FieldValue.where(
            dict(parent_id=self.id, parent_class="Sample")
        )

        to_remove = [
            fv for fv in server_fvs if fv.id not in [_fv.id for _fv in new_fvs]
        ]
        if to_remove:
            warn(
                "Trident tried to save a Sample, but it required FieldValues to be deleted."
            )
            for fv in to_remove:
                fv.parent_id = None
                fv.save()
        self.reload(self.session.utils.json_save("Sample", self.dump()))
        return self

    def available_items(self, object_type_name=None, object_type_id=None):
        query = {"name": object_type_name, "id": object_type_id}
        query = {k: v for k, v in query.items() if v is not None}
        if query == {}:
            return [i for i in self.items if i.location != "deleted"]
        else:
            object_types = self.session.ObjectType.where(query)
            object_type = object_types[0]
            return [
                i
                for i in self.items
                if i.location != "deleted" and i.object_type_id == object_type.id
            ]

    def copy(self):
        """Return a copy of this sample."""
        copied = super().copy()
        copied.anonymize()
        return copied

    def __str__(self):
        return self._to_str("id", "name", "sample_type")


@add_schema
class SampleType(FieldTypeInterface, JSONSaveMixin, ModelBase):
    """A SampleType model."""

    fields = dict(
        samples=HasMany("Sample", "SampleType"),
        field_types=HasMany(
            "FieldType", ref="parent_id", additional_args={"parent_class": "SampleType"}
        ),
        # TODO: operation_type_afts
        # TODO: property_afts
        # TODO: add relationships description
    )

    def field_type(self, name, role=None):
        return super().field_type(name, role)

    @property
    def properties(self):
        props = {}
        for ft in self.field_types:
            if ft.ftype == "sample":
                props[ft.name] = [str(aft) for aft in ft.allowable_field_types]
            else:
                props[ft.name] = ft.ftype
        return props

    def new_sample(self, name, description, project, properties=None):
        if properties is None:
            properties = dict()
        sample = self.session.Sample.new(
            name=name,
            sample_type=self,
            description=description,
            project=project,
            properties=properties,
        )
        return sample

    def _get_update_json(self):
        return self.dump(include=("field_types",))

    def _get_create_json(self):
        return self.dump(include=("field_types",))

    def __str__(self):
        return self._to_str("id", "name")
